@@ -50,7 +50,12 @@ DEFAULT_USER_AGENT = 'keystoneauth1/%s %s %s/%s' % (
 # here and we'll add it to the list as required.
 _LOG_CONTENT_TYPES = set(['application/json'])
 
-_logger = utils.get_logger(__name__)
+_MAX_RETRY_INTERVAL = 60.0
+_EXPONENTIAL_DELAY_START = 0.5
+
+# NOTE(efried): This is defined in oslo_middleware.request_id.INBOUND_HEADER,
+# but it didn't seem worth adding oslo_middleware to requirements just for that
+_REQUEST_ID_HEADER = 'X-Openstack-Request-Id'
 
 
 def _construct_session(session_obj=None):
@@ -81,6 +86,38 @@ def _mv_legacy_headers_for_service(mv_service_type):
     return headers
 
 
+def _sanitize_headers(headers):
+    """Ensure headers are strings and not bytes."""
+    str_dict = {}
+    for k, v in headers.items():
+        if six.PY3:
+            # requests expects headers to be str type in python3, which means
+            # if we get a bytes we need to decode it into a str
+            k = k.decode('ASCII') if isinstance(k, six.binary_type) else k
+            if v is not None:
+                v = v.decode('ASCII') if isinstance(v, six.binary_type) else v
+        else:
+            # requests expects headers to be str type in python2, which means
+            # if we get a unicode we need to encode it to ASCII into a str
+            k = k.encode('ASCII') if isinstance(k, six.text_type) else k
+            if v is not None:
+                v = v.encode('ASCII') if isinstance(v, six.text_type) else v
+        str_dict[k] = v
+    return str_dict
+
+
+class NoOpSemaphore(object):
+    """Empty context manager for use as a default semaphore."""
+
+    def __enter__(self):
+        """Enter the context manager and do nothing."""
+        pass
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        """Exit the context manager and do nothing."""
+        pass
+
+
 class _JSONEncoder(json.JSONEncoder):
 
     def default(self, o):
@@ -159,7 +196,7 @@ def _determine_calling_package():
 
 
 def _determine_user_agent():
-    """Attempt to programatically generate a user agent string.
+    """Attempt to programmatically generate a user agent string.
 
     First, look at the name of the process. Return this unless it is in
     the `ignored` list.  Otherwise, look at the function call stack and
@@ -185,6 +222,47 @@ def _determine_user_agent():
     return name
 
 
+class RequestTiming(object):
+    """Contains timing information for an HTTP interaction."""
+
+    #: HTTP method used for the call (GET, POST, etc)
+    method = None
+
+    #: URL against which the call was made
+    url = None
+
+    #: Elapsed time information
+    elapsed = None  # type: datetime.timedelta
+
+    def __init__(self, method, url, elapsed):
+        self.method = method
+        self.url = url
+        self.elapsed = elapsed
+
+
+class _Retries(object):
+    __slots__ = ('_fixed_delay', '_current')
+
+    def __init__(self, fixed_delay=None):
+        self._fixed_delay = fixed_delay
+        self.reset()
+
+    def __next__(self):
+        value = self._current
+        if not self._fixed_delay:
+            self._current = min(value * 2, _MAX_RETRY_INTERVAL)
+        return value
+
+    def reset(self):
+        if self._fixed_delay:
+            self._current = self._fixed_delay
+        else:
+            self._current = _EXPONENTIAL_DELAY_START
+
+    # Python 2 compatibility
+    next = __next__
+
+
 class Session(object):
     """Maintains client communication state and common functionality.
 
@@ -244,6 +322,14 @@ class Session(object):
                                  that do not share an auth plugin, it can
                                  be provided here. (optional, defaults to
                                  None which means automatically manage)
+    :param bool split_loggers: Split the logging of requests across multiple
+                               loggers instead of just one. Defaults to False.
+    :param bool collect_timing: Whether or not to collect per-method timing
+                                information for each API call. (optional,
+                                defaults to False)
+    :param rate_semaphore: Semaphore to be used to control concurrency
+                           and rate limiting of requests. (optional,
+                           defaults to no concurrency or rate control)
     """
 
     user_agent = None
@@ -256,7 +342,8 @@ class Session(object):
                  cert=None, timeout=None, user_agent=None,
                  redirect=_DEFAULT_REDIRECT_LIMIT, additional_headers=None,
                  app_name=None, app_version=None, additional_user_agent=None,
-                 discovery_cache=None):
+                 discovery_cache=None, split_loggers=None,
+                 collect_timing=False, rate_semaphore=None):
 
         self.auth = auth
         self.session = _construct_session(session)
@@ -273,6 +360,12 @@ class Session(object):
         if discovery_cache is None:
             discovery_cache = {}
         self._discovery_cache = discovery_cache
+        # NOTE(mordred) split_loggers kwarg default is None rather than False
+        # so we can distinguish between the value being set or not.
+        self._split_loggers = split_loggers
+        self._collect_timing = collect_timing
+        self._api_times = []
+        self._rate_semaphore = rate_semaphore or NoOpSemaphore()
 
         if timeout is not None:
             self.timeout = float(timeout)
@@ -318,22 +411,50 @@ class Session(object):
         secure_headers = ('authorization', 'x-auth-token',
                           'x-subject-token', 'x-service-token')
         if header[0].lower() in secure_headers:
-            token_hasher = hashlib.sha1()
+            token_hasher = hashlib.sha256()
             token_hasher.update(header[1].encode('utf-8'))
             token_hash = token_hasher.hexdigest()
-            return (header[0], '{SHA1}%s' % token_hash)
+            return (header[0], '{SHA256}%s' % token_hash)
         return header
 
+    def _get_split_loggers(self, split_loggers):
+        """Get a boolean value from the various argument sources.
+
+        We default split_loggers to None in the kwargs of the Session
+        constructor so we can track set vs. not set. We also accept
+        split_loggers as a parameter in a few other places. In each place
+        we want the parameter, if given by the user, to win.
+        """
+        # None is the default value in each method's kwarg. None means "unset".
+        if split_loggers is None:
+            # If no value was given, try the value set on the instance.
+            split_loggers = self._split_loggers
+        if split_loggers is None:
+            # If neither a value was given on the method, nor a value was given
+            # on the Session constructor, then the value is False.
+            split_loggers = False
+        return split_loggers
+
     def _http_log_request(self, url, method=None, data=None,
                           json=None, headers=None, query_params=None,
-                          logger=_logger):
+                          logger=None, split_loggers=None):
+        string_parts = []
+
+        if self._get_split_loggers(split_loggers):
+            logger = utils.get_logger(__name__ + '.request')
+        else:
+            # Only a single logger was passed in, use string prefixing.
+            string_parts.append('REQ:')
+            if not logger:
+                logger = utils.get_logger(__name__)
+
         if not logger.isEnabledFor(logging.DEBUG):
             # NOTE(morganfainberg): This whole debug section is expensive,
             # there is no need to do the work if we're not going to emit a
             # debug log.
             return
 
-        string_parts = ['REQ: curl -g -i']
+        string_parts.append('curl -g -i')
 
         # NOTE(jamielennox): None means let requests do its default validation
         # so we need to actually check that this is False.
@@ -356,7 +477,8 @@ class Session(object):
             string_parts.append(url)
 
         if headers:
-            for header in headers.items():
+            # Sort headers so that testing can work consistently.
+            for header in sorted(headers.items()):
                 string_parts.append('-H "%s: %s"'
                                     % self._process_header(header))
         if json:
@@ -373,7 +495,18 @@ class Session(object):
 
     def _http_log_response(self, response=None, json=None,
                            status_code=None, headers=None, text=None,
-                           logger=_logger):
+                           logger=None, split_loggers=True):
+        string_parts = []
+        body_parts = []
+        if self._get_split_loggers(split_loggers):
+            logger = utils.get_logger(__name__ + '.response')
+            body_logger = utils.get_logger(__name__ + '.body')
+        else:
+            # Only a single logger was passed in, use string prefixing.
+            string_parts.append('RESP:')
+            body_parts.append('RESP BODY:')
+            body_logger = logger
+
         if not logger.isEnabledFor(logging.DEBUG):
             return
 
@@ -382,6 +515,19 @@ class Session(object):
                 status_code = response.status_code
             if not headers:
                 headers = response.headers
+
+        if status_code:
+            string_parts.append('[%s]' % status_code)
+        if headers:
+            # Sort headers so that testing can work consistently.
+            for header in sorted(headers.items()):
+                string_parts.append('%s: %s' % self._process_header(header))
+        logger.debug(' '.join(string_parts))
+
+        if not body_logger.isEnabledFor(logging.DEBUG):
+            return
+
+        if response is not None:
             if not text:
                 # NOTE(samueldmq): If the response does not provide enough info
                 # about the content type to decide whether it is useful and
@@ -406,17 +552,9 @@ class Session(object):
         if json:
             text = self._json.encode(json)
 
-        string_parts = ['RESP:']
-
-        if status_code:
-            string_parts.append('[%s]' % status_code)
-        if headers:
-            for header in headers.items():
-                string_parts.append('%s: %s' % self._process_header(header))
         if text:
-            string_parts.append('\nRESP BODY: %s\n' % text)
-
-        logger.debug(' '.join(string_parts))
+            body_parts.append(text)
+            body_logger.debug(' '.join(body_parts))
 
     @staticmethod
     def _set_microversion_headers(
@@ -450,8 +588,8 @@ class Session(object):
         # TODO(mordred) cinder uses volume in its microversion header. This
         # logic should be handled in the future by os-service-types but for
         # now hard-code for cinder.
-        if (service_type.startswith('volume')
-                or service_type == 'block-storage'):
+        if (service_type.startswith('volume') or
+                service_type == 'block-storage'):
             service_type = 'volume'
         headers.setdefault('OpenStack-API-Version',
                            '{service_type} {microversion}'.format(
@@ -465,9 +603,12 @@ class Session(object):
                 user_agent=None, redirect=None, authenticated=None,
                 endpoint_filter=None, auth=None, requests_auth=None,
                 raise_exc=True, allow_reauth=True, log=True,
-                endpoint_override=None, connect_retries=0, logger=_logger,
+                endpoint_override=None, connect_retries=0, logger=None,
                 allow=None, client_name=None, client_version=None,
                 microversion=None, microversion_service_type=None,
+                status_code_retries=0, retriable_status_codes=None,
+                rate_semaphore=None, global_request_id=None,
+                connect_retry_delay=None, status_code_retry_delay=None,
                 **kwargs):
         """Send an HTTP request with the specified characteristics.
 
@@ -546,6 +687,28 @@ class Session(object):
                        provided or does not have a service_type, microversion
                        is given and microversion_service_type is not provided,
                        an exception will be raised.
+        :param int status_code_retries: the maximum number of retries that
+                                        should be attempted for retriable
+                                        HTTP status codes (optional, defaults
+                                        to 0 - never retry).
+        :param list retriable_status_codes: list of HTTP status codes that
+                                            should be retried (optional,
+                                            defaults to HTTP 503, has no effect
+                                            when status_code_retries is 0).
+        :param rate_semaphore: Semaphore to be used to control concurrency
+                               and rate limiting of requests. (optional,
+                               defaults to no concurrency or rate control)
+        :param global_request_id: Value for the X-Openstack-Request-Id header.
+        :param float connect_retry_delay: Delay (in seconds) between two
+                                          connect retries (if enabled).
+                                          By default exponential retry starting
+                                          with 0.5 seconds up to a maximum of
+                                          60 seconds is used.
+        :param float status_code_retry_delay: Delay (in seconds) between two
+                                              status code retries (if enabled).
+                                              By default exponential retry
+                                              starting with 0.5 seconds up to
+                                              a maximum of 60 seconds is used.
         :param kwargs: any other parameter that can be passed to
                        :meth:`requests.Session.request` (such as `headers`).
                        Except:
@@ -560,6 +723,17 @@ class Session(object):
 
         :returns: The response to the request.
         """
+        # If a logger is passed in, use it and do not log requests, responses
+        # and bodies separately.
+        if logger:
+            split_loggers = False
+        else:
+            split_loggers = None
+        logger = logger or utils.get_logger(__name__)
+        # HTTP 503 - Service Unavailable
+        retriable_status_codes = retriable_status_codes or [503]
+        rate_semaphore = rate_semaphore or self._rate_semaphore
+
         headers = kwargs.setdefault('headers', dict())
         if microversion:
             self._set_microversion_headers(
@@ -652,9 +826,19 @@ class Session(object):
             headers.setdefault('Content-Type', 'application/json')
             kwargs['data'] = self._json.encode(json)
 
+        if global_request_id is not None:
+            # NOTE(efried): This does *not* setdefault. If a global_request_id
+            # kwarg was explicitly specified, it should override any value
+            # previously configured (e.g. in Adapter.global_request_id).
+            headers[_REQUEST_ID_HEADER] = global_request_id
+
         for k, v in self.additional_headers.items():
             headers.setdefault(k, v)
 
+        # Bug #1766235: some headers may be bytes
+        headers = _sanitize_headers(headers)
+        kwargs['headers'] = headers
+
         kwargs.setdefault('verify', self.verify)
 
         if requests_auth:
@@ -671,7 +855,7 @@ class Session(object):
                                    data=kwargs.get('data'),
                                    headers=headers,
                                    query_params=query_params,
-                                   logger=logger)
+                                   logger=logger, split_loggers=split_loggers)
 
         # Force disable requests redirect handling. We will manage this below.
         kwargs['allow_redirects'] = False
@@ -679,9 +863,15 @@ class Session(object):
         if redirect is None:
             redirect = self.redirect
 
+        connect_retry_delays = _Retries(connect_retry_delay)
+        status_code_retry_delays = _Retries(status_code_retry_delay)
+
         send = functools.partial(self._send_request,
                                  url, method, redirect, log, logger,
-                                 connect_retries)
+                                 split_loggers, connect_retries,
+                                 status_code_retries, retriable_status_codes,
+                                 rate_semaphore, connect_retry_delays,
+                                 status_code_retry_delays)
 
         try:
             connection_params = self.get_auth_connection_params(auth=auth)
@@ -713,13 +903,29 @@ class Session(object):
             request_id = (resp.headers.get('x-openstack-request-id') or
                           resp.headers.get('x-compute-request-id'))
             if request_id:
-                logger.debug('%(method)s call to %(service_name)s for '
-                             '%(url)s used request id '
-                             '%(response_request_id)s',
-                             {'method': resp.request.method,
-                              'service_name': service_name,
-                              'url': resp.url,
-                              'response_request_id': request_id})
+                if self._get_split_loggers(split_loggers):
+                    id_logger = utils.get_logger(__name__ + '.request-id')
+                else:
+                    id_logger = logger
+                if service_name:
+                    id_logger.debug(
+                        '%(method)s call to %(service_name)s for '
+                        '%(url)s used request id '
+                        '%(response_request_id)s', {
+                            'method': resp.request.method,
+                            'service_name': service_name,
+                            'url': resp.url,
+                            'response_request_id': request_id
+                        })
+                else:
+                    id_logger.debug(
+                        '%(method)s call to '
+                        '%(url)s used request id '
+                        '%(response_request_id)s', {
+                            'method': resp.request.method,
+                            'url': resp.url,
+                            'response_request_id': request_id
+                        })
 
         # handle getting a 401 Unauthorized response by invalidating the plugin
         # and then retrying the request. This is only tried once.
@@ -736,10 +942,26 @@ class Session(object):
                          resp.status_code)
             raise exceptions.from_response(resp, method, url)
 
+        if self._collect_timing:
+            for h in resp.history:
+                self._api_times.append(RequestTiming(
+                    method=h.request.method,
+                    url=h.request.url,
+                    elapsed=h.elapsed,
+                ))
+            self._api_times.append(RequestTiming(
+                method=resp.request.method,
+                url=resp.request.url,
+                elapsed=resp.elapsed,
+            ))
+
         return resp
 
-    def _send_request(self, url, method, redirect, log, logger,
-                      connect_retries, connect_retry_delay=0.5, **kwargs):
+    def _send_request(self, url, method, redirect, log, logger, split_loggers,
+                      connect_retries, status_code_retries,
+                      retriable_status_codes, rate_semaphore,
+                      connect_retry_delays, status_code_retry_delays,
+                      **kwargs):
         # NOTE(jamielennox): We handle redirection manually because the
         # requests lib follows some browser patterns where it will redirect
         # POSTs as GETs for certain statuses which is not want we want for an
@@ -753,7 +975,8 @@ class Session(object):
 
         try:
             try:
-                resp = self.session.request(method, url, **kwargs)
+                with rate_semaphore:
+                    resp = self.session.request(method, url, **kwargs)
             except requests.exceptions.SSLError as e:
                 msg = 'SSL exception connecting to %(url)s: %(error)s' % {
                     'url': url, 'error': e}
@@ -779,18 +1002,25 @@ class Session(object):
             if connect_retries <= 0:
                 raise
 
+            delay = next(connect_retry_delays)
             logger.info('Failure: %(e)s. Retrying in %(delay).1fs.',
-                        {'e': e, 'delay': connect_retry_delay})
-            time.sleep(connect_retry_delay)
+                        {'e': e, 'delay': delay})
+            time.sleep(delay)
 
             return self._send_request(
-                url, method, redirect, log, logger,
+                url, method, redirect, log, logger, split_loggers,
+                status_code_retries=status_code_retries,
+                retriable_status_codes=retriable_status_codes,
+                rate_semaphore=rate_semaphore,
                 connect_retries=connect_retries - 1,
-                connect_retry_delay=connect_retry_delay * 2,
+                connect_retry_delays=connect_retry_delays,
+                status_code_retry_delays=status_code_retry_delays,
                 **kwargs)
 
         if log:
-            self._http_log_response(response=resp, logger=logger)
+            self._http_log_response(
+                response=resp, logger=logger,
+                split_loggers=split_loggers)
 
         if resp.status_code in self._REDIRECT_STATUSES:
             # be careful here in python True == 1 and False == 0
@@ -809,17 +1039,45 @@ class Session(object):
                 logger.warning("Failed to redirect request to %s as new "
                                "location was not provided.", resp.url)
             else:
-                # NOTE(jamielennox): We don't pass through connect_retry_delay.
+                # NOTE(jamielennox): We don't keep increasing delays.
                 # This request actually worked so we can reset the delay count.
+                connect_retry_delays.reset()
+                status_code_retry_delays.reset()
                 new_resp = self._send_request(
-                    location, method, redirect, log, logger,
+                    location, method, redirect, log, logger, split_loggers,
+                    rate_semaphore=rate_semaphore,
                     connect_retries=connect_retries,
+                    status_code_retries=status_code_retries,
+                    retriable_status_codes=retriable_status_codes,
+                    connect_retry_delays=connect_retry_delays,
+                    status_code_retry_delays=status_code_retry_delays,
                     **kwargs)
 
                 if not isinstance(new_resp.history, list):
                     new_resp.history = list(new_resp.history)
                 new_resp.history.insert(0, resp)
                 resp = new_resp
+        elif (resp.status_code in retriable_status_codes and
+              status_code_retries > 0):
+
+            delay = next(status_code_retry_delays)
+            logger.info('Retriable status code %(code)s. Retrying in '
+                        '%(delay).1fs.',
+                        {'code': resp.status_code, 'delay': delay})
+            time.sleep(delay)
+
+            # NOTE(jamielennox): We don't keep increasing connection delays.
+            # This request actually worked so we can reset the delay count.
+            connect_retry_delays.reset()
+            return self._send_request(
+                url, method, redirect, log, logger, split_loggers,
+                connect_retries=connect_retries,
+                status_code_retries=status_code_retries - 1,
+                retriable_status_codes=retriable_status_codes,
+                rate_semaphore=rate_semaphore,
+                connect_retry_delays=connect_retry_delays,
+                status_code_retry_delays=status_code_retry_delays,
+                **kwargs)
 
         return resp
 
@@ -974,6 +1232,38 @@ class Session(object):
         auth = self._auth_required(auth, 'determine endpoint URL')
         return auth.get_api_major_version(self, **kwargs)
 
+    def get_all_version_data(self, auth=None, interface='public',
+                             region_name=None, service_type=None,
+                             **kwargs):
+        """Get version data for all services in the catalog.
+
+        :param auth:
+            The auth plugin to use for token. Overrides the plugin on
+            the session. (optional)
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
+        :param interface:
+            Type of endpoint to get version data for. Can be a single value
+            or a list of values. A value of None indicates that all interfaces
+            should be queried. (optional, defaults to public)
+        :param string region_name:
+            Region of endpoints to get version data for. A valueof None
+            indicates that all regions should be queried. (optional, defaults
+            to None)
+        :param string service_type:
+            Limit the version data to a single service. (optional, defaults
+            to None)
+        :returns: A dictionary keyed by region_name with values containing
+            dictionaries keyed by interface with values being a list of
+            `~keystoneauth1.discover.VersionData`.
+        """
+        auth = self._auth_required(auth, 'determine endpoint URL')
+        return auth.get_all_version_data(
+            self,
+            interface=interface,
+            region_name=region_name,
+            service_type=service_type,
+            **kwargs)
+
     def get_auth_connection_params(self, auth=None, **kwargs):
         """Return auth connection params as provided by the auth plugin.
 
@@ -1074,6 +1364,17 @@ class Session(object):
         auth = self._auth_required(auth, 'get project_id')
         return auth.get_project_id(self)
 
+    def get_timings(self):
+        """Return collected API timing information.
+
+        :returns: List of `RequestTiming` objects.
+        """
+        return self._api_times
+
+    def reset_timings(self):
+        """Clear API timing information."""
+        self._api_times = []
+
 
 REQUESTS_VERSION = tuple(int(v) for v in requests.__version__.split('.'))
 
